use aoc2019::intcode;
use std::cmp::Ordering;
use std::collections::HashMap;
use std::error::Error;
use std::fs::File;

use ncurses;

fn main() -> Result<(), Box<dyn Error>> {
    let f = File::open("input13")?;
    let mut computer = intcode::Computer::from(f);
    let mut grid: Grid = HashMap::new();
    while let Some(output) = computer.run_until_n_output(3) {
        let mut output = output.into_iter();
        let xy = XY {
            x: output.next().unwrap(),
            y: output.next().unwrap(),
        };
        let tile = Tile::from(output.next().unwrap());
        grid.insert(xy, tile);
    }

    // part 1
    let blocks = grid.iter().filter(|(_, v)| **v == Tile::Block).count();
    println!("blocks: {}", blocks);

    // part2
    ncurses::initscr();
    ncurses::noecho();
    let mut computer = computer.clone_with_modified_program(vec![(0, 2)]);
    let mut paddlex = 0;
    let mut score = 0;
    while !computer.halted {
        if let Some(output) = &computer.run_until_input() {
            output.chunks(3).for_each(|c| {
                if let &[x, y, value] = c {
                    if x == -1 && y == 0 {
                        score = value;
                    } else {
                        let tile = Tile::from(value);
                        if Tile::from(value) == Tile::HorizontalPaddle {
                            paddlex = x;
                        }
                        ncurses::mvprintw((y + 1) as i32, x as i32, tile.char());
                    }
                }
            });
        }
        let np = next_pos(&mut computer.clone());
        ncurses::mvprintw(
            0,
            0,
            &format!(
                "willx: {:0>2} paddlex: {:0>2} score: {:0>7}",
                np, paddlex, score
            ),
        );
        let joystick = match np.cmp(&paddlex) {
            Ordering::Less => -1,
            Ordering::Equal => 0,
            Ordering::Greater => 1,
        };
        computer.push_input_and_step(joystick);
        ncurses::refresh();
    }
    ncurses::endwin();
    println!("score: {}", score);
    Ok(())
}

fn next_pos(computer: &mut intcode::Computer) -> isize {
    while !computer.halted {
        if let Some(output) = computer.run_until_input() {
            for c in output.chunks(3) {
                if let &[x, y, value] = c {
                    if x == -1 && y == 0 {
                        // skip score
                    } else if Tile::from(value) == Tile::Ball && y == 23 {
                        return x;
                    }
                }
            }
        }
        computer.push_input_and_step(0);
    }
    0
}

#[derive(Eq, PartialEq)]
enum Tile {
    Empty,
    Wall,
    Block,
    HorizontalPaddle,
    Ball,
}

impl From<isize> for Tile {
    fn from(i: isize) -> Self {
        match i {
            0 => Tile::Empty,
            1 => Tile::Wall,
            2 => Tile::Block,
            3 => Tile::HorizontalPaddle,
            4 => Tile::Ball,
            _ => unreachable!("there are only 5 tiles..."),
        }
    }
}

impl Tile {
    fn char(&self) -> &str {
        match self {
            Tile::Empty => " ",
            Tile::Wall => "X",
            Tile::Block => "O",
            Tile::HorizontalPaddle => "-",
            Tile::Ball => "*",
        }
    }
}

#[derive(Eq, PartialEq, Hash)]
struct XY {
    x: isize,
    y: isize,
}

impl From<(isize, isize)> for XY {
    fn from(xy: (isize, isize)) -> Self {
        XY { x: xy.0, y: xy.1 }
    }
}

type Grid = HashMap<XY, Tile>;
